@@ -1,7 +1,7 @@
 /*
  hstr.c     HSTR shell history completion utility
 
- Copyright (C) 2014-2021 Martin Dvorak <martin.dvorak@mindforger.com>
+ Copyright (C) 2014-2023 Martin Dvorak <martin.dvorak@mindforger.com>
 
  Licensed under the Apache License, Version 2.0 (the "License");
  you may not use this file except in compliance with the License.
@@ -137,7 +137,7 @@
 
 // major.minor.revision
 static const char* VERSION_STRING=
-        "hstr version \"2.4.0\" (2021-12-03T21:30:00)"
+        "hstr version \"3.0.0\" (2023-04-15T08:50:00)"
         "\n";
 
 static const char* HSTR_VIEW_LABELS[]={
@@ -157,7 +157,7 @@ static const char* HSTR_CASE_LABELS[]={
         "sensitive"
 };
 
-static const char* INSTALL_BASH_STRING=
+static const char* INSTALL_BASH_CODE_PREFIX=
         "\n# HSTR configuration - add this to ~/.bashrc"
         "\nalias hh=hstr                    # hh to be alias for hstr"
         "\nexport HSTR_CONFIG=hicolor       # get more colors"
@@ -176,96 +176,32 @@ static const char* INSTALL_BASH_STRING=
         //             across different bash sessions.
         //   -c -r ... Forces entire .bash_history to be reloaded (handles history deletes, synchronizes different bash sessions)
         "\n# ensure synchronization between bash memory and history file"
-        "\nexport PROMPT_COMMAND=\"history -a; history -n; ${PROMPT_COMMAND}\""
-        "\n# if this is interactive shell, then bind hstr to Ctrl-r (for Vi mode check doc)"
-#if defined(__MS_WSL__)
-        // IMPROVE commands are NOT executed on return under win10 > consider hstr_utils changes
-        // Script hints:
-        //  {...} is inline group ~ lambda function whose vars are visible to the other commands
-        //   V=$(c) executes commands and stores it to var V
-        "\nfunction hstrwsl {"
-        "\n  offset=${READLINE_POINT}"
-        "\n  READLINE_POINT=0"
-        "\n  { READLINE_LINE=$(</dev/tty hstr ${READLINE_LINE:0:offset} 2>&1 1>&$hstrout); } {hstrout}>&1"
-        "\n  READLINE_POINT=${#READLINE_LINE}"
-        "\n}"
-        "\nif [[ $- =~ .*i.* ]]; then bind -x '\"\\C-r\": \"hstrwsl\"'; fi"
-#elif defined(__CYGWIN__)
-        "\nfunction hstrcygwin {"
-        "\n  offset=${READLINE_POINT}"
-        "\n  READLINE_POINT=0"
-        "\n  { READLINE_LINE=$(</dev/tty hstr ${READLINE_LINE:0:offset} 2>&1 1>&$hstrout); } {hstrout}>&1"
-        "\n  READLINE_POINT=${#READLINE_LINE}"
-        "\n}"
-        "\nif [[ $- =~ .*i.* ]]; then bind -x '\"\\C-r\": \"hstrcygwin\"'; fi"
-#else
-        "\nif [[ $- =~ .*i.* ]]; then bind '\"\\C-r\": \"\\C-a hstr -- \\C-j\"'; fi"
-        "\n# if this is interactive shell, then bind 'kill last command' to Ctrl-x k"
-        "\nif [[ $- =~ .*i.* ]]; then bind '\"\\C-xk\": \"\\C-a hstr -k \\C-j\"'; fi"
-#endif
-        "\n\n";
+        "\nexport PROMPT_COMMAND=\"history -a; history -n; ${PROMPT_COMMAND}\"";
 
 // zsh doc: http://zsh.sourceforge.net/Guide/zshguide.html
-static const char* INSTALL_ZSH_STRING=
+static const char* INSTALL_ZSH_CODE_PREFIX=
         "\n# HSTR configuration - add this to ~/.zshrc"
         "\nalias hh=hstr                    # hh to be alias for hstr"
         "\nsetopt histignorespace           # skip cmds w/ leading space from history"
         // HISTFILE should not be needed - HSTR must work on blank environment as well
         // "\nexport HISTFILE=~/.zsh_history   # ensure history file visibility"
-        "\nexport HSTR_CONFIG=hicolor       # get more colors"
-#if defined(__MS_WSL__)
-        // TODO binding to be rewritten for zsh@WSL as it's done for bash - hstr_winwsl() like function to be implemented to make it work on WSL
-
-        "\n# Function and binding below is bash script that makes command completion work under WSL."
-        "\n# If you can rewrite the function and binding from bash to zsh please send it to martin.dvorak@mindforger.com"
-        "\n# so that I can share it with other users."
-        "\n#function hstr_winwsl {"
-        "\n#  offset=${READLINE_POINT}"
-        "\n#  READLINE_POINT=0"
-        "\n#  { READLINE_LINE=$(</dev/tty hstr ${READLINE_LINE:0:offset} 2>&1 1>&$hstrout); } {hstrout}>&1"
-        "\n#  READLINE_POINT=${#READLINE_LINE}"
-        "\n#}"
-        "\n#bindkey -s \"\\C-r\" \"\\eqhstr_winwsl\\n\""
-        "\n"
-        "\nbindkey -s \"\\C-r\" \"\\eqhstr\\n\"     # bind hstr to Ctrl-r (for Vi mode check doc)"
-#elif defined(__CYGWIN__)
-        // TODO binding to be rewritten for zsh@Cygwin as it's done for bash - hstr_cygwin() like function to be implemented to make it work under Cygwin
-
-        "\n# Function and binding below is bash script that makes command completion work under Cygwin."
-        "\n# If you can rewrite the function and binding from bash to zsh please send it to martin.dvorak@mindforger.com"
-        "\n# so that I can share it with other users."
-        "\n#function hstr_cygwin {"
-        "\n#  offset=${READLINE_POINT}"
-        "\n#  READLINE_POINT=0"
-        "\n#  { READLINE_LINE=$(</dev/tty hstr ${READLINE_LINE:0:offset} 2>&1 1>&$hstrout); } {hstrout}>&1"
-        "\n#  READLINE_POINT=${#READLINE_LINE}"
-        "\n#}"
-        "\n#bindkey -s \"\\C-r\" \"\\eqhstr_cygwin\\n\""
-        "\n"
-        "\nbindkey -s \"\\C-r\" \"\\eqhstr\\n\"     # bind hstr to Ctrl-r (for Vi mode check doc)"
-#else
-        "\nbindkey -s \"\\C-r\" \"\\C-a hstr -- \\C-j\"     # bind hstr to Ctrl-r (for Vi mode check doc)"
-#endif
-        // TODO try variant with args/pars separation
-        //"\nbindkey -s \"\\C-r\" \"\\eqhstr --\\n\"     # bind hstr to Ctrl-r (for Vi mode check doc)"
-        // alternate binding options in zsh:
-        //   bindkey -s '^R' '^Ahstr ^M'
-        //   bindkey -s "\C-r" "\C-a hstr -- \C-j"
-        "\n\n";
+        "\nexport HSTR_CONFIG=hicolor       # get more colors";
 
 static const char* HELP_STRING=
         "Usage: hstr [option] [arg1] [arg2]..."
         "\nShell history suggest box:"
         "\n"
-        "\n  --favorites              -f ... show favorites view"
-        "\n  --kill-last-command      -k ... delete last command in history"
-        "\n  --non-interactive        -n ... print filtered history and exit"
-        "\n  --show-configuration     -s ... show configuration to be added to ~/.bashrc"
-        "\n  --show-zsh-configuration -z ... show zsh configuration to be added to ~/.zshrc"
-        "\n  --show-blacklist         -b ... show commands to skip on history indexation"
-        "\n  --insert-in-terminal=[c] -i ... insert command c in terminal prompt and exit"
-        "\n  --version                -V ... show version details"
-        "\n  --help                   -h ... help"
+        "\n  --favorites               -f ... show favorites view"
+        "\n  --kill-last-command       -k ... delete last command in history"
+        "\n  --non-interactive         -n ... print filtered history and exit"
+        "\n  --show-configuration      -s ... show configuration to be added to ~/.bashrc"
+        "\n  --show-bash-configuration -B ... show bash configuration to be added to ~/.bashrc"
+        "\n  --show-zsh-configuration  -Z ... show zsh configuration to be added to ~/.zshrc"
+        "\n  --show-blacklist          -b ... show commands to skip on history indexation"
+        "\n  --is-tiocsti              -t ... detect whether TIOCSTI is supported and print y or n"
+        "\n  --insert-in-terminal=[c]  -i ... insert command c in terminal prompt and exit"
+        "\n  --version                 -V ... show version details"
+        "\n  --help                    -h ... help"
         "\n"
         "\nReport bugs to martin.dvorak@mindforger.com"
         "\nHome page: https://github.com/dvorka/hstr"
@@ -280,16 +216,18 @@ static const char* LABEL_HELP=
 #define GETOPT_OPTIONAL_ARGUMENT     2
 
 static const struct option long_options[] = {
-        {"favorites",              GETOPT_NO_ARGUMENT, NULL,       'f'},
-        {"kill-last-command",      GETOPT_NO_ARGUMENT, NULL,       'k'},
-        {"version",                GETOPT_NO_ARGUMENT, NULL,       'V'},
-        {"help",                   GETOPT_NO_ARGUMENT, NULL,       'h'},
-        {"non-interactive",        GETOPT_NO_ARGUMENT, NULL,       'n'},
-        {"show-configuration",     GETOPT_NO_ARGUMENT, NULL,       's'},
-        {"show-zsh-configuration", GETOPT_NO_ARGUMENT, NULL,       'z'},
-        {"show-blacklist",         GETOPT_NO_ARGUMENT, NULL,       'b'},
-        {"insert-in-terminal",     GETOPT_REQUIRED_ARGUMENT, NULL, 'i'},
-        {0,                        0,                  NULL,        0 }
+        {"favorites",               GETOPT_NO_ARGUMENT, NULL,       'f'},
+        {"kill-last-command",       GETOPT_NO_ARGUMENT, NULL,       'k'},
+        {"version",                 GETOPT_NO_ARGUMENT, NULL,       'V'},
+        {"help",                    GETOPT_NO_ARGUMENT, NULL,       'h'},
+        {"non-interactive",         GETOPT_NO_ARGUMENT, NULL,       'n'},
+        {"show-configuration",      GETOPT_NO_ARGUMENT, NULL,       's'},
+        {"show-bash-configuration", GETOPT_NO_ARGUMENT, NULL,       'B'},
+        {"show-zsh-configuration",  GETOPT_NO_ARGUMENT, NULL,       'Z'},
+        {"show-blacklist",          GETOPT_NO_ARGUMENT, NULL,       'b'},
+        {"is-tiocsti",              GETOPT_NO_ARGUMENT, NULL,       't'},
+        {"insert-in-terminal",      GETOPT_REQUIRED_ARGUMENT, NULL, 'i'},
+        {0,                         0,                  NULL,        0 }
 };
 
 typedef struct {
@@ -406,6 +344,89 @@ void signal_callback_handler_ctrl_c(int signum)
     }
 }
 
+void print_bash_install_code(void)
+{
+    printf("%s", INSTALL_BASH_CODE_PREFIX);
+
+    if(is_tiocsti) {
+        printf(
+            "\nif [[ $- =~ .*i.* ]]; then bind '\"\\C-r\": \"\\C-a hstr -- \\C-j\"'; fi"
+            "\n# if this is interactive shell, then bind 'kill last command' to Ctrl-x k"
+            "\nif [[ $- =~ .*i.* ]]; then bind '\"\\C-xk\": \"\\C-a hstr -k \\C-j\"'; fi"
+            "\nexport HSTR_TIOCSTI=y"
+        );
+    } else {
+        printf(
+#if defined(__MS_WSL__)
+        "\nfunction hstrwsl {"
+#elif defined(__CYGWIN__)
+        "\nfunction hstrcygwin {"
+#else
+        "\nfunction hstrnotiocsti {"
+#endif
+        "\n    { HSTR_OUT=\"$( { </dev/tty hstr ${READLINE_LINE}; } 2>&1 1>&3 3>&- )\"; } 3>&1;"
+        "\n    READLINE_LINE=\"$(hstr ${READLINE_LINE})\""
+        "\n    READLINE_POINT=${#READLINE_LINE}"
+        "\n}"
+        "\n# if this is interactive shell, then bind hstr to Ctrl-r (for Vi mode check doc)"
+
+#if defined(__MS_WSL__)
+        "\nif [[ $- =~ .*i.* ]]; then bind -x '\"\\C-r\": \"hstrwsl\"'; fi"
+#elif defined(__CYGWIN__)
+        "\nif [[ $- =~ .*i.* ]]; then bind -x '\"\\C-r\": \"hstrcygwin\"'; fi"
+#else
+        "\nif [[ $- =~ .*i.* ]]; then bind -x '\"\\C-r\": \"hstrnotiocsti\"'; fi"
+#endif
+        );
+        printf("\nexport HSTR_TIOCSTI=n");
+    }
+
+    printf("\n\n");
+}
+
+void print_zsh_install_code(void)
+{
+    printf("%s", INSTALL_ZSH_CODE_PREFIX);
+
+    if(is_tiocsti) {
+        printf(
+            "\nbindkey -s \"\\C-r\" \"\\C-a hstr -- \\C-j\"     # bind hstr to Ctrl-r (for Vi mode check doc)"
+            "\nexport HSTR_TIOCSTI=y"
+        );
+    } else {
+        printf(
+#if defined(__MS_WSL__)
+        "\nhstr_wsl() {"
+#elif defined(__CYGWIN__)
+        "\nhstr_cygwin() {"
+#else
+        "\nhstr_no_tiocsti() {"
+#endif
+
+        "\n    zle -I"
+        "\n    { HSTR_OUT=\"$( { </dev/tty hstr ${BUFFER}; } 2>&1 1>&3 3>&- )\"; } 3>&1;"
+        "\n    BUFFER=\"${HSTR_OUT}\""
+        "\n    CURSOR=${#BUFFER}"
+        "\n    zle redisplay"
+        "\n}"
+
+#if defined(__MS_WSL__)
+        "\nzle -N hstr_wsl"
+        "\nbindkey '\\C-r' hstr_wsl"
+#elif defined(__CYGWIN__)
+        "\nzle -N hstr_cygwin"
+        "\nbindkey '\\C-r' hstr_cygwin"
+#else
+        "\nzle -N hstr_no_tiocsti"
+        "\nbindkey '\\C-r' hstr_no_tiocsti"
+#endif
+        );
+        printf("\nexport HSTR_TIOCSTI=n");
+    }
+
+    printf("\n\n");
+}
+
 unsigned recalculate_max_history_items(void)
 {
     int n = getmaxy(stdscr);
@@ -742,7 +763,7 @@ void print_cmd_added_favorite_label(const char* cmd)
         color_attr_on(COLOR_PAIR(HSTR_COLOR_INFO));
         color_attr_on(A_BOLD);
     }
-    mvprintw(hstr->promptYNotification, 0, screenLine);
+    mvprintw(hstr->promptYNotification, 0, "%s", screenLine);
     if(hstr->theme & HSTR_THEME_COLOR) {
         color_attr_off(A_BOLD);
         color_attr_on(COLOR_PAIR(HSTR_COLOR_NORMAL));
@@ -1098,7 +1119,7 @@ char* hstr_print_selection(unsigned maxHistoryItems, char* pattern)
                 if(hstr->matching==HSTR_MATCH_REGEXP) {
                     start=hstr->selectionRegexpMatch[i].rm_so;
                     count=hstr->selectionRegexpMatch[i].rm_eo-start;
-                    if(count>CMDLINE_LNG) {
+                    if(count>=CMDLINE_LNG) {
                         count=CMDLINE_LNG-1;
                     }
                     strncpy(screenLine,
@@ -1207,6 +1228,7 @@ void stdout_history_and_return(void)
 }
 
 // IMPROVE hstr doesn't have to be passed as parameter - it's global static
+// TODO rename method _
 char* getResultFromSelection(int selectionCursorPosition, Hstr* hstr, char* result) {
     if (hstr->promptBottom) {
         result=hstr->selection[hstr->promptItems-selectionCursorPosition-1];
@@ -1272,8 +1294,8 @@ void loop_to_select(void)
     char pattern[SELECTION_PREFIX_MAX_LNG];
     pattern[0]=0;
     // TODO this is too late! > don't render twice
-    // TODO overflow
-    strcpy(pattern, hstr->cmdline);
+    strncpy(pattern, hstr->cmdline, SELECTION_PREFIX_MAX_LNG-1);
+    pattern[SELECTION_PREFIX_MAX_LNG-1]=0;
 
     while (!done) {
         maxHistoryItems=recalculate_max_history_items();
@@ -1625,6 +1647,7 @@ void loop_to_select(void)
             fill_terminal_input("\"", FALSE);
         }
         if(executeResult) {
+            // TODO w/o TIOCSTI the command is NOT executed, just shown in the prompt
             fill_terminal_input("\n", FALSE);
         }
     }
@@ -1670,7 +1693,7 @@ void hstr_interactive(void)
 void hstr_getopt(int argc, char **argv)
 {
     int option_index = 0;
-    int option = getopt_long(argc, argv, "fkVhnszbi", long_options, &option_index);
+    int option = getopt_long(argc, argv, "fkVhnsBZbti", long_options, &option_index);
     if(option != -1) {
         switch(option) {
         case 'f':
@@ -1700,20 +1723,32 @@ void hstr_getopt(int argc, char **argv)
             printf("%s", VERSION_STRING);
             hstr_exit(EXIT_SUCCESS);
             break;
+        case 't':
+            if(is_tiocsti) {
+                printf("y\n");
+            } else {
+                printf("n\n");
+            }
+            hstr_exit(EXIT_SUCCESS);
+            break;
         case 'h':
             printf("%s", HELP_STRING);
             hstr_exit(EXIT_SUCCESS);
             break;
-        case 'z':
-            printf("%s", INSTALL_ZSH_STRING);
+        case 'B':
+            print_bash_install_code();
+            hstr_exit(EXIT_SUCCESS);
+            break;
+        case 'Z':
+            print_zsh_install_code();
             hstr_exit(EXIT_SUCCESS);
             break;
         case 's':
             // ZSH_VERSION is not exported by zsh > detected by parent process name
-            if(isZshParentShell()) {
-                printf("%s", INSTALL_ZSH_STRING);
+            if(is_zsh_parent_shell()) {
+                print_zsh_install_code();
             } else {
-                printf("%s", INSTALL_BASH_STRING);
+                print_bash_install_code();
             }
             hstr_exit(EXIT_SUCCESS);
             break;
@@ -1733,6 +1768,13 @@ int hstr_main(int argc, char* argv[])
 {
     setlocale(LC_ALL, "");
 
+    // initialize global TIOCSTI indicator
+#ifdef DEBUG_NO_TIOCSTI
+    is_tiocsti = false;
+#else
+    is_tiocsti = is_tiocsti_supported();
+#endif
+
     hstr=malloc(sizeof(Hstr));
     hstr_init();
 
